﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.ComponentModel;
using System.Collections.ObjectModel;
using Buzz.MachineInterface;
using BuzzGUI.Interfaces;
using BuzzGUI.Common;
using System.Threading;
using System.Collections;

namespace WDE.Freezer2
{
    enum FreezerState
    {
        Through = 0,       
        Play,
        Record
    };
    
	[MachineDecl(Name = "WDE Freezer 2", ShortName = "Freezer 2", Author = "WDE", MaxTracks = 1)]
	public class Freezer2 : IBuzzMachine, INotifyPropertyChanged
	{
		IBuzzMachineHost host;
        SampleData sampleData;
        Hashtable inputVolumes;
        int currentSlot = 0;
        bool recording = false;

        public Freezer2(IBuzzMachineHost host)
		{
			this.host = host;
            sampleData = new SampleData();
            inputVolumes = new Hashtable();
		}

        int freezerState = (int)FreezerState.Through;

        [ParameterDecl(MaxValue = 2, DefValue = 0, ValueDescriptions = new String[] { "Through", "Play", "Record" })]
        public int Mode { get { return freezerState; }
            set
            {                

                switch ((int)value)
                {
                    case (int)FreezerState.Record:                                            
                        Application.Current.Dispatcher.BeginInvoke(new Action(() =>
                        {
                            MuteInputs(false);
                            sampleData.ZeroBuffers();
                            host.MasterInfo.PosInTick = 0;
                            recording = true;
                        }));
                        break;
                    case (int)FreezerState.Through:
                        Application.Current.Dispatcher.BeginInvoke(new Action(() =>
	                    {                 
                            MuteInputs(false);
                            recording = false;
                        }));                        
                        break;
                    case (int)FreezerState.Play:                      
                        Application.Current.Dispatcher.BeginInvoke(new Action(() =>
	                    {                 
                            MuteInputs(true);                            
                            if ((autoCopy == 1) && (recording == true))
                            {
                                saveToWavetable(currentSlot);                                
                            }
                            host.MasterInfo.PosInTick = 0;
                            recording = false;
                        }));
                        break;                    
                }
                freezerState = (int)value;
            }
        }

        int autoCopy = 0;

        [ParameterDecl(MaxValue = 1, DefValue = 0, ValueDescriptions = new String[] { "False", "True" })]
        public int AutoCopy
        {
            get { return autoCopy; }
            set
            {
                autoCopy = value;
            }
        }

        public void Stop()
        {
            if (freezerState == (int)FreezerState.Record)
            {
                foreach (IParameterGroup paramGroup in host.Machine.ParameterGroups)
                    foreach (IParameter param in paramGroup.Parameters)
                    {                        
                        if (param.Name == "Mode")
                        {
                            param.SetValue(0,(int)FreezerState.Play);
                        }
                    }
            }
        }

        public bool Work(Sample[] output, Sample[] input, int numsamples, WorkModes mode)
        {
 
            
            if (mode == WorkModes.WM_NOIO)
                return false;
            if (mode == WorkModes.WM_READ)                        // <thru>
                return true;   

            bool ret = true; //Create sound

            switch (freezerState)
            {
                case (int)FreezerState.Record:
                    {
                        if (mode == WorkModes.WM_WRITE)
                            return false;
                                                
                        if (Global.Buzz.Song.Buzz.Playing || Global.Buzz.Song.Buzz.Recording)
                        {

                            int playPosition = Global.Buzz.Song.PlayPosition * host.MasterInfo.SamplesPerTick + host.MasterInfo.PosInTick;                                                      
                            
                            sampleData.AddSampleBuf(input, numsamples, playPosition);
                            for (int i = 0; i < numsamples; i++)
                            {
                                output[i] = input[i];
                            }
                        }
                    }
                    break;

                case (int)FreezerState.Play:
                    {

                        if (Global.Buzz.Song.Buzz.Playing || Global.Buzz.Song.Buzz.Recording)
                        {
                            // Why is host.MasterInfo.PosInTick non-zero when you start playing the song from position 0???
                            int playPosition = Global.Buzz.Song.PlayPosition * host.MasterInfo.SamplesPerTick + host.MasterInfo.PosInTick;
                            
                            sampleData.GetSampleBuf(ref output, numsamples, playPosition);
                            ret = true;
                        }
                        else
                        {
                            ret = false;
                        }
                    }
                    break;
                case (int)FreezerState.Through:
                    {
                        for (int i = 0; i < numsamples; i++)
                        {
                            output[i] = input[i];
                        }
                        if (mode == WorkModes.WM_WRITE)
                            return false;
                    }
                    break;
            }
                
            return ret;
        }

        public void ZeroAmp(bool mute, IMachineConnection conn)
        {
            if (mute)
            {   
                inputVolumes[conn.Source.Name] = conn.Amp;
                conn.Amp = 0;
            }
            else
            {
                try
                {
                    int amp = (int)inputVolumes[conn.Source.Name];
                    conn.Amp = amp;
                }
                catch (Exception)
                {
                }                
            }
        }

        public void MuteInputs(bool mute)
        {
            if (host.Machine != null)
            {
                foreach (IMachineConnection macCon in host.Machine.Inputs)
                {
                    IMachine mac = macCon.Source;                    
                    mac.IsMuted = mute;
                    ZeroAmp(mute, macCon);
                }
            }
        }
		// actual machine ends here. the stuff below demonstrates some other features of the api.
	
		public class State : INotifyPropertyChanged
		{
			public State() { text = "here is state"; }	// NOTE: parameterless constructor is required by the xml serializer

			string text;
			public string Text 
			{
				get { return text; }
				set
				{
					text = value;
					if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs("Text"));
					// NOTE: the INotifyPropertyChanged stuff is only used for data binding in the GUI in this demo. it is not required by the serializer.
				}
			}
            
			public event PropertyChangedEventHandler PropertyChanged;            
		}

		State machineState = new State();
		public State MachineState			// a property called 'MachineState' gets automatically saved in songs and presets
		{
			get { return machineState; }
			set
			{
				machineState = value;
				if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs("MachineState"));
			}
		}

        public IEnumerable<IMenuItem> Commands
        {
            get
            {

                yield return new MenuItemVM()
                {
                    Text = "Help",
                    Command = new SimpleCommand()
                    {
                        CanExecuteDelegate = p => true,
                        ExecuteDelegate = p => MessageBox.Show("Modes\n\nThrough: Input audio piped to output.\nPlay: Play recorded audio.\nRecord: Record inputs, automatically switch to play mode when stopped.")
                    }
                };

                yield return new MenuItemVM()
                {
                    Text = "About...",
                    Command = new SimpleCommand()
                    {
                        CanExecuteDelegate = p => true,
                        ExecuteDelegate = p => MessageBox.Show("Version 1.0.1 (C) 2017 WDE")
                    }
                };
            }
        }

        public void saveToWavetable(int slot)
        {
            var wt = host.Machine.Graph.Buzz.Song.Wavetable;
            WaveFormat wf;
            wf = WaveFormat.Float32;
            
            // write to wavetable
            int rootnote = BuzzNote.FromMIDINote(48);
            wt.AllocateWave(slot,
                                "",
                                "Freezer",
                                (int)(sampleData.GetBuffer().Length),
                                wf,
                                true,
                                rootnote,
                                false,
                                true);
            IWaveLayer layer = wt.Waves[slot].Layers.Last();
            layer.SampleRate = Global.Buzz.SelectedAudioDriverSampleRate;

            float[] tmpBuffer = new float[sampleData.GetBuffer().Length];
            for (int i = 0; i < sampleData.GetBuffer().Length; i++)
            {
                tmpBuffer[i] = sampleData.GetBuffer()[i] / 32768.0f;
            }

            layer.SetDataAsFloat(tmpBuffer, 0, 2, 0, 0, (int)sampleData.GetBuffer().Length / 2); // Left
            layer.SetDataAsFloat(tmpBuffer, 1, 2, 1, 0, (int)sampleData.GetBuffer().Length / 2); // Right
            
            layer.InvalidateData();
        }

		public event PropertyChangedEventHandler PropertyChanged;
        
        internal void slotUpdated(int decAgain)
        {
            currentSlot = decAgain;
        }
    }

	public class MachineGUIFactory : IMachineGUIFactory { public IMachineGUI CreateGUI(IMachineGUIHost host) { return new Freezer2GUI(); } }
	public class Freezer2GUI : UserControl, IMachineGUI
	{
		IMachine machine;
		Freezer2 freezer2Machine;
		TextBox tb;		
        Button bt;
        ComboBox comboWT;
        //Canvas visuals;

		// view model for machine list box items
		public class MachineVM
		{
			public IMachine Machine { get; private set; }
			public MachineVM(IMachine m) { Machine = m; }
			public override string ToString() { return Machine.Name; }
		}
		
		public ObservableCollection<MachineVM> Machines { get; private set; }
		
		public IMachine Machine
		{
			get { return machine; }
			set
			{
				if (machine != null)
				{
					BindingOperations.ClearBinding(tb, TextBox.TextProperty);
				}

				machine = value;
                
				if (machine != null)
				{
					freezer2Machine = (Freezer2)machine.ManagedMachine;
                    
					tb.SetBinding(TextBox.TextProperty, new Binding("MachineState.Text") { Source = freezer2Machine, Mode = BindingMode.TwoWay, UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged });
				}
			}
		}


		public Freezer2GUI()
		{
			Machines = new ObservableCollection<MachineVM>();

            bt = new Button() { Margin = new Thickness(0, 0, 0, 4), Height = 30 };
            bt.Content = new TextBlock() { Text = "Copy Recorded Audio to Wavetable" };
            TextBlock textBlock = new TextBlock() { Text = "Select Wavetable Slot", Margin = new Thickness(0, 0, 0, 4) };
            bt.Click += new RoutedEventHandler(bt_Click);
            
			tb = new TextBox() { Margin = new Thickness(0, 0, 0, 4), AllowDrop = true };
			            
			var sp = new StackPanel();
            
            sp.Children.Add(bt);
            sp.Children.Add(textBlock);

            this.Content = sp;
            comboWT = new ComboBox() { Margin = new Thickness(0, 0, 0, 4) };
            for (int i = 1; i < 201; i++)
            {
                comboWT.Items.Add(new TextBlock() {Text= i.ToString("X")});                
            }
            comboWT.SelectedIndex = 0;
            comboWT.SelectionChanged += new SelectionChangedEventHandler(comboWT_SelectionChanged);

            sp.Children.Add(comboWT);
			// drag and drop example
			tb.PreviewDragEnter += (sender, e) => { e.Effects = DragDropEffects.Copy; e.Handled = true; };
			tb.PreviewDragOver += (sender, e) => { e.Effects = DragDropEffects.Copy; e.Handled = true; };
			tb.Drop += (sender, e) => { tb.Text = "drop"; e.Handled = true; };			
		}

        void comboWT_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            string hex = ((TextBlock)comboWT.SelectedItem).Text;
            int decAgain = int.Parse(hex, System.Globalization.NumberStyles.HexNumber) - 1;
            if (freezer2Machine != null)
                this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Normal,
                 new Action(
                   delegate()
                   {
                       freezer2Machine.slotUpdated(decAgain);
                   }
               )); 
        }

        void bt_Click(object sender, RoutedEventArgs e)
        {
            string hex = ((TextBlock)comboWT.SelectedItem).Text;
            string message = "Are you sure you want to copy recorded audio to Wavetable slot " + hex + "?";
            string caption = "Copy audio to wavetable?";
            MessageBoxButton buttons = MessageBoxButton.YesNo;

            MessageBoxResult result = MessageBox.Show(message, caption, buttons);
            if (result == MessageBoxResult.Yes)
            {
                int decAgain = int.Parse(hex, System.Globalization.NumberStyles.HexNumber) -1;
                if (freezer2Machine != null)
                    this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Normal,
                     new Action(
                       delegate()
                       {
                           freezer2Machine.saveToWavetable(decAgain);
                       }
                   )); 
                    
            }            
        }
	}
}
